from typing import Iterable, Tuple, Dict, Callable
import torch.distributions as d

import torch as t

from evostrat import Population, Individual


class MultivariateNormalPopulation(Population):
    """
    A distribution over individuals whose parameters are sampled from a multivariate normal with a full covariance matrix.
    """

    def __init__(self,
                 individual_parameter_shapes: Dict[str, t.Size],
                 individual_constructor: Callable[[Dict[str, t.Tensor]], Individual]
                 ):
        """
        A distribution over individuals whose parameters are sampled from a multivariate normal with a full covariance matrix.

        :param individual_parameter_shapes: The shapes of the parameters of an individual
        :param individual_constructor: Constructs individuals from parameters
        """
        n_params = sum([s.numel() for s in individual_parameter_shapes.values()])
        self.means = t.zeros((n_params,), dtype=t.float32, requires_grad=True)
        self.log_stds = (-6 * t.ones((n_params, n_params)) + 6 * t.eye(n_params, n_params)).clone().detach().requires_grad_(True)
        self.shapes = individual_parameter_shapes
        self.constructor = individual_constructor

    def _to_shapes(self, flat_params) -> Dict[str, t.Tensor]:
        params = {}
        i = 0
        for k, s in self.shapes.items():
            params[k] = flat_params[i:i + s.numel()].reshape(s)
            i += s.numel()
        return params

    def parameters(self) -> Iterable[t.Tensor]:
        return [self.means, self.log_stds]

    def sample(self, n) -> Iterable[Tuple[Individual, t.Tensor]]:
        dist = d.MultivariateNormal(loc=self.means, scale_tril=t.exp(self.log_stds).tril())
        with t.no_grad():
            samples = dist.sample((n,))
        log_probs = dist.log_prob(samples)
        individuals = [self.constructor(self._to_shapes(s)) for s in samples]

        return zip(individuals, log_probs.unbind(0))



import torch as t
from typing import Iterable, Callable, Tuple
from torch.multiprocessing import Pool

from evostrat import Individual


class Population:
    """
    A parameterized distribution over individuals.

    Meant to be optimized with `torch.optim` optimizers, as follows:

    pop = PopulationImpl(...)
    optim = torch.optim.Adam(pop.parameters())
    for i in range(N):
        optim.zero_grads()
        pop.fitness_grads(n_samples=200)
        optim.step()
    """

    def parameters(self) -> Iterable[t.Tensor]:
        """
        :return: The parameters of this population distribution.
        """

        raise NotImplementedError

    def sample(self, n) -> Iterable[Tuple[Individual, t.Tensor]]:
        """
        Sample n individuals and compute their log probabilities. The log probability computation MUST be differentiable.

        :param n: How many individuals to sample
        :return: n individuals and their log probability of being sampled: [(ind_1, log_prob_1), ..., (ind_n, log_prob_n)]
        """
        raise NotImplementedError

    def fitness_grads(
            self,
            n_samples: int,
            pool: Pool = None,
            fitness_shaping_fn: Callable[[Iterable[float]], Iterable[float]] = lambda x: x
    ):
        """
        Computes the (approximate) gradients of the expected fitness of the population.

        Uses torch autodiff to compute the gradients. The Individual.fitness does NOT need to be differentiable,
        but the log probability computations in Population.sample MUST be.

        :param n_samples: How many individuals to sample to approximate the gradient
        :param pool: Optional process pool to use when computing the fitness of the sampled individuals.
        :param fitness_shaping_fn: Optional function to modify the fitness, e.g. normalization, etc. Input is a list of n raw fitness floats. Output must also be n floats.
        :return: A (n,) tensor containing the raw fitness (before fitness_shaping_fn) for the n individuals.
        """

        individuals, log_probs = zip(*self.sample(n_samples))
        assert all(lp.ndim == 0 and lp.isfinite() and lp.grad_fn is not None for lp in log_probs), "log_probs must be differentiable finite scalars"

        if pool is not None:
            raw_fitness = pool.map(_fitness_fn_no_grad, individuals)
        else:
            raw_fitness = list(map(_fitness_fn_no_grad, individuals))

        fitness = fitness_shaping_fn(raw_fitness)

        t.mean(t.stack([(-ind_fitness * log_prob) for log_prob, ind_fitness in zip(log_probs, fitness)])).backward()

        return t.tensor(raw_fitness)


def _fitness_fn_no_grad(ind: Individual):
    with t.no_grad():
        return ind.fitness()



import tqdm as tqdm
from torch.multiprocessing import Pool
from torch.optim import Adam

from evostrat import compute_centered_ranks, MultivariateNormalPopulation
from evostrat.examples.lunar_lander import LunarLander

if __name__ == '__main__':
    """
    Lunar landers weights and biases are drawn from a multivariate normal distribution with learned means and a learned covariance matrix.
    
    This is similar to CMA-ES [1].
    [1] - Hansen, Nikolaus, and Andreas Ostermeier. "Completely derandomized self-adaptation in evolution strategies." Evolutionary computation 9.2 (2001): 159-195.
    """
    param_shapes = {k: v.shape for k, v in LunarLander().get_params().items()}
    population = MultivariateNormalPopulation(param_shapes, LunarLander.from_params)

    learning_rate = 0.1
    iterations = 1000
    pop_size = 200

    optim = Adam(population.parameters(), lr=learning_rate)
    pbar = tqdm.tqdm(range(iterations))
    pool = Pool()

    for _ in pbar:
        optim.zero_grad()
        raw_fit = population.fitness_grads(pop_size, pool, compute_centered_ranks)
        optim.step()
        pbar.set_description("fit avg: %0.3f, std: %0.3f" % (raw_fit.mean().item(), raw_fit.std().item()))
        if raw_fit.mean() > 200:
            print("Solved.")
            break

    pool.close()



from typing import Dict

import gym
from torch import nn
import torch as t

from evostrat import Individual


class LunarLander(Individual):
    """
    A lunar lander controlled by a feedforward policy network
    """

    def __init__(self):
        self.net = nn.Sequential(
            nn.Linear(8, 32), nn.Tanh(),
            nn.Linear(32, 4), nn.Softmax(dim=0)
        )

    @staticmethod
    def from_params(params: Dict[str, t.Tensor]) -> 'LunarLander':
        agent = LunarLander()
        agent.net.load_state_dict(params)
        return agent

    def fitness(self, render=False) -> float:
        env = gym.make("LunarLander-v2")
        obs = env.reset()
        done = False
        r_tot = 0
        while not done:
            action = self.action(obs)
            obs, r, done, _ = env.step(action)
            r_tot += r
            if render:
                env.render()

        env.close()
        return r_tot

    def get_params(self) -> Dict[str, t.Tensor]:
        return self.net.state_dict()

    def action(self, obs):
        with t.no_grad():
            return t.argmax(self.net(t.tensor(obs, dtype=t.float32))).item()



class Individual:
    """
    An individual which has a fitness.
    """

    def fitness(self) -> float:
        """
        :return: The fitness of the individual. Does NOT need to be differentiable.
        """
        raise NotImplementedError



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='evostrat',
    version='1.1.1',
    author="Rasmus Berg Palm",
    author_email="rasmusbergpalm@gmail.com",
    description="A library that makes Evolutionary Strategies (ES) simple to use.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/rasmusbergpalm/evostrat",
    packages=setuptools.find_packages(),
    install_requires=[
        "torch>=1.6.0"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ]
)
